# -*- coding: utf-8 -*-
import re

import six

from .utils import Importer, create_path_function


class BaseProperty(object):
    def get(self, instance):
        raise NotImplementedError

    @property
    def raw_value(self):
        raise NotImplementedError

    @classmethod
    def build(cls, name, value):
        raise NotImplementedError


class ValueProperty(BaseProperty):
    def __init__(self, value):
        self.value = value

    def get(self, instance):
        return self.value

    @property
    def raw_value(self):
        return self.value

    @classmethod
    def build(cls, name, value):
        if isinstance(value, BaseProperty):
            return value, True
        return cls(value), True


class InterpolationProperty(BaseProperty):
    var_re = re.compile('\{[ ]*(\w+)[ ]*\}')

    def __init__(self, name, template, var_names):
        self.name = name
        self.template = template
        self.var_names = var_names

    def call_callables(self, instance, name):
        attr = instance[name]
        func = getattr(attr, '__call__', lambda: attr)
        return func()

    def get(self, instance):
        values = dict([
            (vname, self.call_callables(instance, vname))
            for vname in self.var_names
        ])
        return self.template.format(**values)

    @property
    def raw_value(self):
        return self.template

    @classmethod
    def build(cls, name, value):
        if isinstance(value, six.string_types):
            var_names = cls.var_re.findall(value)
            if var_names:
                return cls(name, value, var_names), True
        return None, False


class ImporterProperty(BaseProperty):
    def __init__(self, importer):
        self.importer = importer

    def get(self, instance):
        return self.importer.import_it()

    @property
    def raw_value(self):
        return self.importer

    @classmethod
    def build(cls, name, value):
        if isinstance(value, Importer):
            return cls(value), True
        return None, False


class RawProperty(BaseProperty, six.text_type):
    def get(self, instance):
        return self

    @property
    def raw_value(self):
        return self

    @classmethod
    def build(cls, name, value):
        if isinstance(value, cls):
            return value, True
        return None, False


class LazyProperty(BaseProperty):
    def __init__(self, get):
        self.__get = get

    def get(self, instance):
        return self.__get(instance)

    @property
    def raw_value(self):
        return self.__get

    @classmethod
    def build(cls, name, value):
        if isinstance(value, cls):
            return value, True
        return None, False


class LazyRootpathProperty(BaseProperty):
    def __init__(self, *path):
        self.path_args = path
        self.__rootpath = None

    def get(self, instance):
        rootpath_function = self.__rootpath
        if not rootpath_function:
            rootpath_function = create_path_function(instance['__rootpath__'])
            self.__rootpath = rootpath_function
        return rootpath_function(*self.path_args)

    @property
    def raw_value(self):
        return self.path_args

    @classmethod
    def build(cls, name, value):
        if isinstance(value, cls):
            return value, True
        return None, False



# -*- coding: utf-8 -*-
import os
import sys

from .utils import create_path_function, extrabuiltins, syspath, Importer, split_filenames
from .collection import Collection, RawProperty, LazyProperty, Module, LazyRootpathProperty
from .sources import ModuleSource, EnvironmentVariableSource, INISource


class Loader(object):
    collection = Collection
    lazyimport = Importer
    lazy = LazyProperty
    raw = RawProperty
    rootpath = LazyRootpathProperty

    def __init__(self, file=None, syspaths=None):
        self._file = file or './_current_directory_'
        self._syspaths = syspaths or ['.']
        self._rootpath = create_path_function(self._file)

    # processing configuration
    def _get_syspath(self):
        return [self._rootpath(*s.split(os.sep)) for s in self._syspaths]

    def _get_extrabuildins(self):
        return {
            'confy': self
        }
    # end

    # api
    def merge(self, *sources):
        def do_merge(context):
            for s in sources:
                context = s.load(context)
            return context

        with syspath(self._get_syspath()):
            with extrabuiltins(self._get_extrabuildins()):
                context = do_merge({})

        return Collection.collectionize(context, defaults={
            '__rootpath__': self._file,
        })

    def define_module(self, name, sources):
        sys.modules[name] = Module(name, self._file, self.merge(*sources))

    def from_modules(self, *files, **kwargs):
        return ModuleSource(
            names=split_filenames(files, abspath=self._rootpath, ext=kwargs.get('ext', 'py')),
            silent=kwargs.get('silent'),
        )

    def from_environ_vars(self, names, silent=False):
        return EnvironmentVariableSource(
            names=names,
            silent=silent,
            environ=os.environ,
        )

    def from_ini(self, *files, **kwargs):
        return INISource(
            names=split_filenames(files, abspath=self._rootpath, ext=kwargs.get('ext', 'ini')),
            silent=kwargs.get('silent')
        )
    # end



# -*- coding: utf-8 -*-
from types import ModuleType
from collections import MutableMapping, Mapping

import six

from .properties import (
    LazyProperty, RawProperty, InterpolationProperty,
    ImporterProperty, ValueProperty, LazyRootpathProperty,
)


class Collection(object):

    # For extendibility purposes we might want to change / add some properties
    property_classes = [
        LazyRootpathProperty,
        LazyProperty,
        RawProperty,
        InterpolationProperty,
        ImporterProperty,
        ValueProperty,
    ]

    def _choose_property(self, name, value):
        for property_cls in self.property_classes:
            prop, success = property_cls.build(name, value)
            if success:
                return prop
        raise ValueError("No property for: %s | %s" % (name, value))
    # end


    # collectionize
    @classmethod
    def collectionize(cls, adict, defaults=None):
        """ Change a dict into a collection. All inner dictionaries
        are changed into collections as well.

        :param adict: dictionary (collections.Mapping) alike object
        :param defaults: default attributes that will be added to all
        new collection objects
        """

        def new(adict):
            c = cls(defaults or {})
            c.update(adict)
            return c

        def convert(adict):
            for key in list(adict.keys()):
                value = adict[key]
                if isinstance(value, Mapping):
                    adict[key] = new(convert(value))
            return new(adict)

        return convert(adict)
    # end


    def __init__(self, *args, **kwargs):
        dict_data_key = "_%s__data" % self.__class__.__name__
        self.__dict__.update({
            dict_data_key: {},
        })
        self.update(*args, **kwargs)


    # Protected general read / write methods
    def __set(self, name, value):
        self.__data[name] = self._choose_property(name, value)

    def __property(self, name):
        try:
            return self.__data[name]
        except KeyError:
            raise KeyError("Collection has no such key: %s" % name)

    def __get(self, name):
        prop = self.__property(name)
        return prop.get(self)

    def __raw(self, name):
        prop = self.__property(name)
        return prop.raw_value
    # end


    # [] set, get & del protocol
    def __setitem__(self, name, value):
        self.__set(name, value)

    def __getitem__(self, name):
        return self.__get(name)

    def __delitem__(self, name):
        del self.__data[name]
    # end


    # support for get attribute protocol
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        return self.__get(name)

    def __setattr__(self, name, value):
        self.__set(name, value)
    # end


    def pop(self, key, **kwargs):
        try:
            prop = self.__data.pop(key, **kwargs)
        except KeyError:
            raise KeyError("Collection has no such key: %s" % key)
        return prop.get(self)

    def popitem(self):
        try:
            key = next(iter(self.__data))
        except StopIteration:
            raise KeyError
        value = self.__get(key)
        del self.__data[key]
        return key, value

    def update(*args, **kwds):
        args_len = len(args)

        if args_len > 2:
            raise TypeError(
                "update() takes at most 2 positional arguments ({0} given)".format(args_len)
            )
        elif not args:
            raise TypeError("update() takes at least 1 argument (0 given)")

        self = args[0]
        other = args[1] if args_len >= 2 else ()

        if isinstance(other, Mapping):
            for key in other:
                self.__set(key, other[key])
        elif hasattr(other, "keys"):
            for key in other.keys():
                self.__set(key, other[key])
        else:
            for key, value in other:
                self.__set(key, value)
        for key, value in kwds.items():
            self.__set(key, value)

    def setdefault(self, key, default=None):
        try:
            return self.__get(key)
        except KeyError:
            self.__set(key, default)
        return default

    def clear(self):
        self.__data.clear()

    def keys(self):
        return self.__data.keys()

    def items(self):
        return [(name, self.__get(name)) for name in self.__data]

    def values(self):
        return [self.__get(key) for key in self]

    def iterkeys(self):
        return iter(self.__data.keys())

    def itervalues(self):
        for key in self:
            yield self.__get(key)

    def iteritems(self):
        for key in self:
            yield (key, self.__get(key))

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key):
        return key in self.__data

    def get(self, key, default=None):
        try:
            return self.__get(key)
        except KeyError:
            return default

    def __eq__(self, other):
        if not isinstance(other, Mapping):
            return False
        return dict(self.items()) == dict(other.items())

    def __ne__(self, other):
        return not (self == other)

    def __len__(self):
        return len(self.__data)

    # Extra *Collection* methods
    def properties(self):
        return [(name, self.__data[name]) for name in self.keys()]

    def extend(self, *args, **kwargs):
        attrs = Collection(self.properties())
        attrs.update(*args, **kwargs)
        return attrs

    def raw(self, name):
        return self.__raw(name)

    def raw_items(self):
        return [(name, self.__raw(name)) for name in self.keys()]

    def raw_values(self):
        return [self.__raw(name) for name in self.keys()]
    # end

    def __str__(self):
        return six.u('<Collection: %s>' % [
            key for key in self.keys() if not key.startswith('_')
        ])

    __unicode__ = __repr__ = __str__


MutableMapping.register(Collection)


class Module(ModuleType):
    def __init__(self, module_name, filepath, collection):
        super(Module, self).__init__(module_name)
        self.__file__ = filepath
        for name, value in collection.items():
            setattr(self, name, value)

    def __getitem__(self, name):
        return getattr(self, name)



# -*- coding: utf-8 -*-
VERSION = '0.3.3'


from .api import Loader
from .collection import Module, Collection
from .properties import (
    BaseProperty, ValueProperty, InterpolationProperty, ImporterProperty,
    RawProperty, LazyProperty,
)

loader = Loader
collection = Loader.collection

lazyimport = Loader.lazyimport
lazy = Loader.lazy
raw = Loader.raw
rootpath = Loader.rootpath



# -*- coding: utf-8 -*-
import os

from .utils import execfile


class BaseSource(object):
    pass


class ModuleSource(BaseSource):
    def __init__(self, names, silent=False):
        self.names = names
        self.silent = silent

    def _before_execfiles(self, context):
        return context

    def _execfile(self, absolute_path, context, silent=False):
        try:
            execfile(absolute_path, global_vars=context)
        except IOError:
            if not silent:
                raise
        return context

    def _after_execfile(self, context, module_name):
        return context

    def load(self, context):
        context = self._before_execfiles(context)

        for m in self.names:
            context = self._execfile(m, context, silent=self.silent)
            context = self._after_execfile(context, m)

        return context


class EnvironmentVariableSource(BaseSource):
    def __init__(self, names, silent=False, environ=None):
        self.names = names
        self.silent = silent
        self.environ = environ or os.environ

    def load(self, context):
        data = {}
        for vardef in self.names:
            if ":" in vardef:
                env_name, setting_name = vardef.split(":")
            else:
                env_name = setting_name = vardef
            try:
                data[setting_name] = self.environ[env_name]
            except KeyError:
                if not self.silent:
                    raise
        context.update(data)
        return context


class INISource(BaseSource):
    def __init__(self, names, sections=None, silent=False):
        self.names = names
        self.silent = silent
        self.sections = sections  # None means ALL

    def load(self, context):
        import configparser

        names = []
        for name in self.names:
            if not os.path.exists(name) and not self.silent:
                raise IOError("No such file or directory: '%s'" % name)
            names.append(name)

        cp = configparser.ConfigParser()
        cp.read(names)

        sections = self.sections or cp.sections()
        try:
            config_as_dict = dict([
                (section_name, dict(cp[section_name]))
                for section_name in sections
            ])
            context.update(config_as_dict)
        except KeyError:  # cp[<key>] raises KeyError not NoSectionError
            if not self.silent:
                raise

        return context



# -*- coding: utf-8 -*-
import sys
from os.path import join, dirname, abspath

import six

from .packages import importlib


def create_path_function(thefile):
    root = abspath(join(dirname(abspath(thefile))))
    return lambda *a: abspath(join(root, *a))


class extrabuiltins(object):
    def __init__(self, things, builtins=None):
        self.things = things or {}
        self.builtins = builtins or six.moves.builtins

    def extend(self):
        for name, thing in self.things.items():
            setattr(self.builtins, name, thing)

    def cleanup(self):
        for name in self.things:
            delattr(self.builtins, name)

    def __enter__(self):
        self.extend()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()


class syspath(object):
    def __init__(self, paths, sys_path=None):
        self.paths = paths or []
        self.sys_path = sys.path if sys_path == None else sys_path

    def extend(self):
        for path in reversed(self.paths):
            self.sys_path.insert(0, path)

    def cleanup(self):
        for path in reversed(self.paths):
            self.sys_path.remove(path)

    def __enter__(self):
        self.extend()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()



def execfile(file_path, global_vars=None, local_vars=None):
    with open(file_path) as f:
        code = compile(f.read(), file_path, 'exec')
        six.exec_(code, global_vars, local_vars)
    return global_vars, local_vars


class Importer(object):
    def __init__(self, path):
        self.__path = path
        self.__importeditem = None

    def import_it(self):
        if not self.__importeditem:
            if '.' in self.__path:
                path_to_item, last_item = self.__path.rsplit('.', 1)
            else:
                path_to_item, last_item = self.__path, None

            item = importlib.import_module(path_to_item)

            if last_item is None:
                self.__importeditem = item
                return item

            try:
                self.__importeditem = getattr(item, last_item)
            except AttributeError as e:
                raise ImportError("Module '%s': => %s" % (path_to_item, six.text_type(e)))
        return self.__importeditem


def smart_ext(name, ext=None):
    """
    Smartly add extention to the file name

    Usage::

        >>> smart_ext("config")
        "config"
        >>> smart_ext("config.py")
        "config.py"
        >>> smart_ext("config", ext="py")
        "config.py"
        >>> smart_ext("config.py", ext="py")
        "config.py"

    """
    if ext is None:
        return name

    ext = ".%s" % ext if not ext.startswith(".") else ext
    name = name.strip()
    existing_ext = name.rsplit(".", 1)[-1]

    if existing_ext == name:  # no ext acctually
        return "%s%s" % (name, ext)

    return name


def split_filenames(names, abspath=None, ext=None):
    abspath = abspath or (lambda x: x)
    splited = []
    for name in names:
        splited.extend(
            [abspath(smart_ext(s, ext=ext)) for s in name.split(',')]
        )
    return splited



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


import confy


setup(
    name='confy',
    version=confy.VERSION,
    description='',
    author='Jakub Janoszek',
    author_email='kuba.janoszek@gmail.com',
    url='https://pypi.python.org/pypi/confy',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
    ],
    packages=find_packages(exclude=("tests*", "examples*")),
    include_package_data = True,
    zip_safe=False,
    install_requires = [
        'six==1.3.0',
    ],
)

# Usage of setup.py:
# $> python setup.py register             # registering package on PYPI
# $> python setup.py build sdist upload   # build, make source dist and upload to PYPI
